/*
Copyright (C) 2022 Christoph Boelitz

E-mail: tools@dosforum.de
Project page: http://tools.dosforum.de/

This program is free software as long as its used privately.
Commercial use, as unlikely as it may seem, is hereby prohibited.
The source code may be modified and distributed as long
as my contact information is left in place.

Let me know, if you have suggestions or see something that can
be done in a better way, so I might learn from it. Thanks!
*/

#include "ide.h"

unsigned int ide_swap_bytes(unsigned int w)
{
  char b1 = w >> 8;
  char b2 = w - (b1 << 8);
  return (b2 << 8)+ b1;
}

int ide_check_if_cdrom(drive_info* di)
{
  //Check, if device speaks ATAPI

  if
  (
    (di->general_config & IDE_ATAPI_GEN_CFG_WORD_PROTOTYPE)
    != IDE_ATAPI_GEN_CFG_WORD_PROTOTYPE_ATAPI
  ) //11000000
    return IDE_ERR_NOT_ATAPI_DEV;

  //Check, if device is a CD-ROM device
  if
  (
    (di->general_config & IDE_ATAPI_GEN_CFG_WORD_DEVTYPE)
    != IDE_ATAPI_GEN_CFG_WORD_DEVTYPE_CDROM
  )
    return IDE_ERR_NOT_A_CDROM;

  //Check, if device accepts 12-byte ATAPI packets
  //(All CD-ROMs according to spec)

  if
  (
    (di->general_config & IDE_ATAPI_GEN_CFG_WORD_PKTSIZE)
    != IDE_ATAPI_GEN_CFG_WORD_PKTSIZE_12B
  )
    return IDE_ERR_WRONG_PKT_SIZE;

  return IDE_ERR_SUCCESS;
}

int ide_get_atapi_info(unsigned int port, unsigned int device, drive_info* di)
{
  /*
    1. Send IDENTIFY DRIVE ATA command; Drive will
       abort and reveal ATAPI signature.
    2. Read ATAPI signature.
    3. Send IDENTIFY DEVICE ATAPI command; Drive will
       store information in sector buffer.
    4. Read information from sector buffer and store
       in drive_info struct.

    Note: According to spec the IDENTIFY DRIVE(!) ATA command shall be
    aborted by CD-ROM drives, so as not to confuse them with ATA drives.
  */

  int ret; //Used for various return values
  unsigned int id_drive_data[256];
  unsigned int ide_cyl_high, ide_cyl_low;
  int i;
  unsigned int error;

  //1
  ret = ide_send_ata_cmd(port, device, IDE_IDENTIFY_DRIVE, 1, 1, 0);

  if (ret != IDE_ERR_SUCCESS)
  {
    _enable();
    return ret;
  }

  ret = ide_check_error_condition(port, &error);

  //At this point, if the device is an ATAPI device, it should have aborted
  //the previous command.

  if (ret != IDE_ERR_COMMAND_ABORTED)
  {
    _enable();
    return IDE_ERR_IDDRIVE_NOT_ABORTED;
  }

  //2
  ide_cyl_high = inp(port+IDE_CYL_HIGH);
  ide_cyl_low = inp(port+IDE_CYL_LOW);

  //Re-enable interrupts as they were disabled by ide_send_ata_cmd().
  _enable();

  if
  (
    (ide_cyl_high == ATAPI_SIG_CYL_HIGH) &&
    (ide_cyl_low  == ATAPI_SIG_CYL_LOW)
  )
    return IDE_ERR_NO_ATAPI_SIG;

  //3 (Spec says: "Shall be accepted regardless of DRDY state")
  ret = ide_send_ata_cmd(port, device, IDE_ATAPI_IDENTIFY_DEVICE, 1, 0, 1);

  if (ret != IDE_ERR_SUCCESS)
  {
    _enable();
    return ret;
  }

  //4
  for (i=0; i<256; i++)
    id_drive_data[i] = inpw(port+IDE_DATA);

  _enable();

  //Copy model number (offset 27-46)
  for (i=27;i<=46;i++)
  {
    di->model_number[(i-27)*2] = (char)(id_drive_data[i] >> 8);
    di->model_number[(i-27)*2+1] = (char)id_drive_data[i];
  }

  di->model_number[sizeof(di->model_number)-1] = '\0';

  //Copy firmware revision (offset 23-26)
  for (i=23;i<=26;i++)
  {
    di->firmware_revision[(i-23)*2]
      = (char)(id_drive_data[i] >> 8);
    di->firmware_revision[(i-23)*2+1]
      = (char)id_drive_data[i];
  }

  di->firmware_revision[sizeof(di->firmware_revision)-1] = '\0';

  di->general_config = id_drive_data[0]; //One word, offset 0

  return IDE_ERR_SUCCESS;
}

int ide_poll_register(unsigned int port, unsigned int mask, unsigned int state)
{
  //Polls IDE register until bits set in mask match their value in state
  //Specify bits that should be tested in mask by setting them high and
  //their respective value in state.

  //Note: Be sure to specify the register port, not the IDE base port.

  unsigned int reg_bits;
  int i;

  i=0;
  reg_bits = inp(port);

  while ((reg_bits & mask) != state)
  {
    delay(100);


    i++;

    if (i > 40) // try ~4 seconds...
      return IDE_ERR_TIMEOUT;

    reg_bits = inp(port);
  }

  return IDE_ERR_SUCCESS;
}

int ide_send_ata_cmd(unsigned int port, unsigned int device,
  unsigned int opcode, int wait_busy, int wait_ready, int wait_drq)
{
  /*
    Sends ATAPI command:
      1. Poll status until BSY=0 (optional via wait_busy)
      2. Disable interrupts (clear bit NIEN)
      3. Poll status until DRDY=1 (optional via wait_ready)
      4. select device (master/slave)
      Only for PACKET command:
        5. Set IEN=0 in device control register (0=enable ints)
      6. send ATAPI command (opcode)
      7. Wait a bit
      8. Poll status until DRQ=1 (optional via wait_drq)

    Note: This command disables interrupts.
  */

  int ret; //Used for various return values

  //1
  if (wait_busy)
  {
    ret = ide_poll_register(port+IDE_STATUS, IDE_STATUS_BUSY, 0);

    if (ret != IDE_ERR_SUCCESS)
      return IDE_ERR_WAITBUSY;
  }

  //2
  _disable();

  //3
  if (wait_ready)
  {
    ret = ide_poll_register(port+IDE_STATUS, IDE_STATUS_DRDY, IDE_STATUS_DRDY);

    if (ret != IDE_ERR_SUCCESS)
      return IDE_ERR_WAITDRDY;
  }

  //4
  outp(port+IDE_DRIVE_HEAD, device);

  //5
  if (opcode == IDE_ATAPI_PACKET_COMMAND)
    outp(port+IDE_DEVCTRL, IDE_DEVCTRL_NIEN); //0000 0010

  //6
  outp(port+IDE_STATUS, opcode);

  //7 For ATAPI packet command we should wait 400ns, but we'll wait a bit longer
  delay(5); // Probably overkill, but will do the trick.

  //8
  if (wait_drq)
  {
    ret = ide_poll_register(port+IDE_STATUS, IDE_STATUS_DRQ, IDE_STATUS_DRQ);

    if (ret != IDE_ERR_SUCCESS)
      return IDE_ERR_WAITDRQ;
  }

  return IDE_ERR_SUCCESS;
}

int ide_check_error_condition(unsigned int port, unsigned int *r_error)
{
  //Checks IDE error regsiter for error and returns type of error.

  //If this function is called right after ide_send_ata_cmd() it seems we
  //have to give the controller some time to validate the error register,
  //so we wait for BSY=0 here.

  int ret;
  unsigned int status_bits;

  *r_error = 0;

  ret = ide_poll_register(port+IDE_STATUS, IDE_STATUS_BUSY, 0);

  if (ret != IDE_ERR_SUCCESS)
    return IDE_ERR_WAITBUSY;

  status_bits = inp(port+IDE_STATUS);

  if ((status_bits & IDE_STATUS_ERR) == 0)
    return IDE_ERR_SUCCESS;

  //Else: There has been an error; Retrieve error bits from error register
  //      to investigate further. Also, since the ERR bit has been set in
  //      status register, we can assume that the error register is valid.

  *r_error = inp(port+IDE_ERROR);

  if (((*r_error) & IDE_ERROR_ABRT) == IDE_ERROR_ABRT)
    return IDE_ERR_COMMAND_ABORTED;

  //Spec defines more errors, but they are not of interest here.
  return IDE_ERR_UNKNOWN_ERROR;
}

void ide_atapi_packet_set_cd_speed(unsigned int packet[], unsigned int speed)
{
  //Fills an array of words that represent the
  //SET CD SPEED ATAPI command packet

  packet[0] = IDE_ATAPI_PKT_SET_CDROM_SPEED;
  packet[1] = speed==0 ? 0xffff : ide_swap_bytes(speed); //0xffff = max speed
  packet[2] = 0;
  packet[3] = 0;
  packet[4] = 0;
  packet[5] = 0;
}

int ide_atapi_soft_reset(unsigned int port, int device)
{
  //Sends soft-reset command to IDE device

  int ret;
  unsigned int error;

  //Spec says, shall be accepted when BSY or DRQ is high, DRDY is low
  ret = ide_send_ata_cmd(port, device, IDE_ATAPI_SOFTRESET, 0, 0, 0);

  if (ret != IDE_ERR_SUCCESS)
  {
    _enable();
    return ret;
  }

  //If supported this command will not end in an error condition, otherwise
  //it will probably result in command aborted state.
  ret = ide_check_error_condition(port, &error);

  _enable();

  if (ret == IDE_ERR_COMMAND_ABORTED)
    return IDE_ERR_ATAPI_SRST_ABORTED;

  if (ret != IDE_ERR_SUCCESS)
    return ret;

  return IDE_ERR_SUCCESS;
}

int ide_atapi_send_packet_cmd(unsigned int port, unsigned int device,
  unsigned int packet[], unsigned int *r_error)
{
  int ret; //Used for various return values
  int i;

  ret = ide_send_ata_cmd(port, device, IDE_ATAPI_PACKET_COMMAND, 1, 1, 1);

  if (ret != IDE_ERR_SUCCESS)
  {
    _enable();
    return ret;
  }

  delay(5);

  //Wait for BSY=0
  ret = ide_poll_register(port+IDE_STATUS, IDE_STATUS_BUSY, 0);

  if (ret != IDE_ERR_SUCCESS)
    return IDE_ERR_WAITBUSY;

  //Sending the packet command is done by 6 word writes. These
  //make up the 12 bytes of the command's size.

  for(i=0; i<6; i++)
  {
    //outportw(), because we send a word (16 bit)
    outpw(port+IDE_DATA, packet[i]);

    //Wait one I/O cycle by reading the alternate status register.
    inp(port+IDE_ALT_STATUS);
  }

  //Check alternative status register one more time to waste one i/o cycle.
  inp(port+IDE_ALT_STATUS);

  ret = ide_check_error_condition(port, r_error);

  _enable();

  return ret;
}

